@@ -3,6 +3,7 @@
  * This file is part of GtkSourceView
  *
  * Copyright (C) 2009 - Jesse van den Kieboom
+ * Copyright (C) 2013 - Sébastien Wilmet
  *
  * gtksourceview is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -33,17 +34,16 @@
 #include "gtksourcecompletionwordslibrary.h"
 #include "gtksourcecompletionwordsbuffer.h"
 #include "gtksourcecompletionwordsutils.h"
+#include "gtksourceview/gtksource.h"
+#include "gtksourceview/gtksourceview-typebuiltins.h"
+#include "gtksourceview/gtksourceview-i18n.h"
 
-#include <gtksourceview/gtksourcecompletion.h>
-#include <gtksourceview/gtksourceview-i18n.h>
 #include <string.h>
 
 #define GTK_SOURCE_COMPLETION_WORDS_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE((object), GTK_SOURCE_TYPE_COMPLETION_WORDS, GtkSourceCompletionWordsPrivate))
 
 #define BUFFER_KEY "GtkSourceCompletionWordsBufferKey"
 
-#define GET_WORDS_BUFFER(buf) (((BufferBinding *)g_object_get_data(G_OBJECT(buf), BUFFER_KEY))->buffer)
-
 enum
 {
 	PROP_0,
@@ -54,7 +54,8 @@ enum
 	PROP_SCAN_BATCH_SIZE,
 	PROP_MINIMUM_WORD_SIZE,
 	PROP_INTERACTIVE_DELAY,
-	PROP_PRIORITY
+	PROP_PRIORITY,
+	PROP_ACTIVATION
 };
 
 struct _GtkSourceCompletionWordsPrivate
@@ -80,6 +81,7 @@ struct _GtkSourceCompletionWordsPrivate
 
 	gint interactive_delay;
 	gint priority;
+	GtkSourceCompletionActivation activation;
 };
 
 typedef struct
@@ -187,51 +189,24 @@ add_in_idle (GtkSourceCompletionWords *words)
 	return !finished;
 }
 
-static gboolean
-valid_word_char (gunichar ch,
-                 gpointer data)
-{
-	return g_unichar_isprint (ch) && (ch == '_' || g_unichar_isalnum (ch));
-}
-
-static gboolean
-valid_start_char (gunichar ch,
-                  gpointer data)
-{
-	return !g_unichar_isdigit (ch);
-}
-
 static gchar *
-get_word_at_iter (GtkTextIter    *iter,
-                  CharacterCheck  valid,
-                  CharacterCheck  valid_start,
-                  gpointer        data)
+get_word_at_iter (GtkTextIter *iter)
 {
-	GtkTextIter end = *iter;
+	GtkTextBuffer *buffer;
+	GtkTextIter start_line;
+	gchar *line_text;
+	gchar *word;
 
-	if (!gtk_source_completion_words_utils_forward_word_end (iter, valid, data) ||
-	    !gtk_text_iter_equal (iter, &end))
-	{
-		return NULL;
-	}
+	buffer = gtk_text_iter_get_buffer (iter);
+	start_line = *iter;
+	gtk_text_iter_set_line_offset (&start_line, 0);
 
-	if (!gtk_source_completion_words_utils_backward_word_start (iter,
-	                                                            valid,
-	                                                            valid_start,
-	                                                            data))
-	{
-		return NULL;
-	}
+	line_text = gtk_text_buffer_get_text (buffer, &start_line, iter, FALSE);
 
-	if (gtk_text_iter_equal (iter, &end))
-	{
-		return NULL;
-	}
-	else
-	{
-		return gtk_text_iter_get_text (iter, &end);
-	}
+	word = _gtk_source_completion_words_utils_get_end_word (line_text);
 
+	g_free (line_text);
+	return word;
 }
 
 static void
@@ -239,24 +214,22 @@ gtk_source_completion_words_populate (GtkSourceCompletionProvider *provider,
                                       GtkSourceCompletionContext  *context)
 {
 	GtkSourceCompletionWords *words = GTK_SOURCE_COMPLETION_WORDS (provider);
+	GtkSourceCompletionActivation activation;
 	GtkTextIter iter;
 	gchar *word;
-	GtkTextBuffer *buffer;
-	GtkSourceCompletionWordsBuffer *buf;
 
 	gtk_source_completion_context_get_iter (context, &iter);
-	buffer = gtk_text_iter_get_buffer (&iter);
 
 	g_free (words->priv->word);
 	words->priv->word = NULL;
 
-	word = get_word_at_iter (&iter,
-	                         valid_word_char,
-	                         valid_start_char,
-	                         words);
+	word = get_word_at_iter (&iter);
+
+	activation = gtk_source_completion_context_get_activation (context);
 
 	if (word == NULL ||
-	    g_utf8_strlen (word, -1) < words->priv->minimum_word_size)
+	    (activation == GTK_SOURCE_COMPLETION_ACTIVATION_INTERACTIVE &&
+	     g_utf8_strlen (word, -1) < words->priv->minimum_word_size))
 	{
 		g_free (word);
 		gtk_source_completion_context_add_proposals (context,
@@ -277,11 +250,6 @@ gtk_source_completion_words_populate (GtkSourceCompletionProvider *provider,
 	words->priv->word = word;
 	words->priv->word_len = strlen (word);
 
-	buf = GET_WORDS_BUFFER (buffer);
-	gtk_text_buffer_move_mark (buffer,
-	                           gtk_source_completion_words_buffer_get_mark (buf),
-	                           &iter);
-
 	/* Do first right now */
 	if (add_in_idle (words))
 	{
@@ -388,6 +356,10 @@ gtk_source_completion_words_set_property (GObject      *object,
 			self->priv->priority = g_value_get_int (value);
 			break;
 
+		case PROP_ACTIVATION:
+			self->priv->activation = g_value_get_flags (value);
+			break;
+
 		default:
 			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 			break;
@@ -432,6 +404,10 @@ gtk_source_completion_words_get_property (GObject    *object,
 			g_value_set_int (value, self->priv->priority);
 			break;
 
+		case PROP_ACTIVATION:
+			g_value_set_flags (value, self->priv->activation);
+			break;
+
 		default:
 			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 			break;
@@ -514,6 +490,23 @@ gtk_source_completion_words_class_init (GtkSourceCompletionWordsClass *klass)
 	                                                   0,
 	                                                   G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
 
+	/**
+	 * GtkSourceCompletionWords:activation:
+	 *
+	 * The type of activation.
+	 *
+	 * Since: 3.10
+	 */
+	g_object_class_install_property (object_class,
+	                                 PROP_ACTIVATION,
+					 g_param_spec_flags ("activation",
+							     _("Activation"),
+							     _("The type of activation"),
+							     GTK_SOURCE_TYPE_COMPLETION_ACTIVATION,
+							     GTK_SOURCE_COMPLETION_ACTIVATION_INTERACTIVE |
+							     GTK_SOURCE_COMPLETION_ACTIVATION_USER_REQUESTED,
+							     G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
+
 	g_type_class_add_private (object_class, sizeof(GtkSourceCompletionWordsPrivate));
 }
 
@@ -523,18 +516,18 @@ gtk_source_completion_words_get_start_iter (GtkSourceCompletionProvider *provide
                                             GtkSourceCompletionProposal *proposal,
                                             GtkTextIter                 *iter)
 {
-	GtkTextBuffer *buffer;
-	GtkSourceCompletionWordsBuffer *buf;
-	GtkTextIter it;
+	gchar *word;
+	glong nb_chars;
 
-	gtk_source_completion_context_get_iter (context, &it);
+	gtk_source_completion_context_get_iter (context, iter);
 
-	buffer = gtk_text_iter_get_buffer (&it);
-	buf = GET_WORDS_BUFFER (buffer);
+	word = get_word_at_iter (iter);
+	g_return_val_if_fail (word != NULL, FALSE);
 
-	gtk_text_buffer_get_iter_at_mark (buffer,
-	                                  iter,
-	                                  gtk_source_completion_words_buffer_get_mark (buf));
+	nb_chars = g_utf8_strlen (word, -1);
+	gtk_text_iter_backward_chars (iter, nb_chars);
+
+	g_free (word);
 	return TRUE;
 }
 
@@ -550,6 +543,12 @@ gtk_source_completion_words_get_priority (GtkSourceCompletionProvider *provider)
 	return GTK_SOURCE_COMPLETION_WORDS (provider)->priv->priority;
 }
 
+static GtkSourceCompletionActivation
+gtk_source_completion_words_get_activation (GtkSourceCompletionProvider *provider)
+{
+	return GTK_SOURCE_COMPLETION_WORDS (provider)->priv->activation;
+}
+
 static void
 gtk_source_completion_words_iface_init (GtkSourceCompletionProviderIface *iface)
 {
@@ -559,6 +558,7 @@ gtk_source_completion_words_iface_init (GtkSourceCompletionProviderIface *iface)
 	iface->get_start_iter = gtk_source_completion_words_get_start_iter;
 	iface->get_interactive_delay = gtk_source_completion_words_get_interactive_delay;
 	iface->get_priority = gtk_source_completion_words_get_priority;
+	iface->get_activation = gtk_source_completion_words_get_activation;
 }
 
 static void
