@@ -81,14 +81,12 @@
 #include "gtksourcecompletioninfo.h"
 #include "gtksourcecompletionproposal.h"
 #include "gtksourcecompletionprovider.h"
+#include "gtksourcecompletioncontainer.h"
 #include "gtksourcebuffer.h"
 #include "gtksourceview.h"
 #include "gtksourceview-marshal.h"
 #include "gtksourceview-i18n.h"
 
-#define WINDOW_WIDTH 350
-#define WINDOW_HEIGHT 200
-
 #define GTK_SOURCE_COMPLETION_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE ((object),\
 						  GTK_SOURCE_TYPE_COMPLETION,           \
 						  GtkSourceCompletionPrivate))
@@ -196,27 +194,31 @@ scroll_to_iter (GtkSourceCompletion *completion,
 	GtkTreePath *path;
 	GtkTreeIter prev_iter = *iter;
 
-	/* If we want to scroll to the first proposal of a provider, it's better
-	 * to show the header too, if there is a header. */
+	path = gtk_tree_model_get_path (GTK_TREE_MODEL (completion->priv->model_proposals),
+					iter);
+
+	gtk_tree_view_scroll_to_cell (completion->priv->tree_view_proposals,
+				      path, NULL,
+				      FALSE, 0, 0);
+	gtk_tree_path_free (path);
+
 	if (gtk_source_completion_model_iter_previous (completion->priv->model_proposals, &prev_iter) &&
 	    gtk_source_completion_model_iter_is_header (completion->priv->model_proposals, &prev_iter))
 	{
+		/* If we want to scroll to the first proposal of a provider,
+		 * it's better to show the header too, if there is a header.
+		 * We first scroll to the proposal, and then to the
+		 * header, so we are sure that the proposal is visible.
+		 */
+
 		path = gtk_tree_model_get_path (GTK_TREE_MODEL (completion->priv->model_proposals),
 						&prev_iter);
-	}
-	else
-	{
-		path = gtk_tree_model_get_path (GTK_TREE_MODEL (completion->priv->model_proposals),
-						iter);
-	}
 
-	gtk_tree_view_scroll_to_cell (completion->priv->tree_view_proposals,
-				      path,
-				      NULL,
-				      FALSE,
-				      0,
-				      0);
-	gtk_tree_path_free (path);
+		gtk_tree_view_scroll_to_cell (completion->priv->tree_view_proposals,
+					      path, NULL,
+					      FALSE, 0, 0);
+		gtk_tree_path_free (path);
+	}
 }
 
 /* Returns %TRUE if a proposal is selected.
@@ -1028,6 +1030,7 @@ gtk_source_completion_move_page (GtkSourceCompletion *completion,
 	}
 
 	gtk_tree_view_set_model (completion->priv->tree_view_proposals, NULL);
+	gtk_tree_view_columns_autosize (completion->priv->tree_view_proposals);
 
 	gtk_source_completion_model_set_visible_providers (completion->priv->model_proposals,
 							   visible_providers);
@@ -1281,6 +1284,8 @@ update_completion (GtkSourceCompletion        *completion,
 
 	/* Create a new CompletionModel */
 	gtk_tree_view_set_model (completion->priv->tree_view_proposals, NULL);
+	gtk_tree_view_columns_autosize (completion->priv->tree_view_proposals);
+
 	replace_model (completion);
 
 	for (item = providers_copy; item != NULL; item = g_list_next (item))
@@ -1467,11 +1472,6 @@ populating_done (GtkSourceCompletion        *completion,
 	update_selection_label (completion);
 	update_bottom_bar_visibility (completion);
 
-	if (!gtk_widget_get_visible (GTK_WIDGET (completion->priv->main_window)))
-	{
-		g_signal_emit (completion, signals[SHOW], 0);
-	}
-
 	if (!check_first_selected (completion))
 	{
 		/* Update the window position only if the first proposal is not
@@ -1480,6 +1480,11 @@ populating_done (GtkSourceCompletion        *completion,
 		 */
 		update_window_position (completion);
 	}
+
+	if (!gtk_widget_get_visible (GTK_WIDGET (completion->priv->main_window)))
+	{
+		g_signal_emit (completion, signals[SHOW], 0);
+	}
 }
 
 static void
@@ -2250,14 +2255,15 @@ init_main_window (GtkSourceCompletion *completion,
 	gtk_window_set_attached_to (GTK_WINDOW (completion->priv->main_window),
 				    GTK_WIDGET (completion->priv->view));
 
-	gtk_widget_set_size_request (GTK_WIDGET (completion->priv->main_window),
-	                             WINDOW_WIDTH,
-	                             WINDOW_HEIGHT);
+	g_signal_connect (completion->priv->main_window,
+			  "configure-event",
+			  G_CALLBACK (gtk_source_completion_configure_event),
+			  completion);
 
-	g_signal_connect_after (completion->priv->main_window,
-				"configure-event",
-				G_CALLBACK (gtk_source_completion_configure_event),
-				completion);
+	g_signal_connect_swapped (completion->priv->main_window,
+				  "size-allocate",
+				  G_CALLBACK (update_window_position),
+				  completion);
 
 	g_signal_connect (completion->priv->main_window,
 			  "delete-event",
@@ -2300,21 +2306,33 @@ init_info_window (GtkSourceCompletion *completion)
 static void
 gtk_source_completion_init (GtkSourceCompletion *completion)
 {
+	GError *error = NULL;
 	GtkBuilder *builder = gtk_builder_new ();
+	GtkSourceCompletionContainer *container = _gtk_source_completion_container_new ();
+	g_object_ref_sink (container);
 
 	completion->priv = GTK_SOURCE_COMPLETION_GET_PRIVATE (completion);
 
 	gtk_builder_set_translation_domain (builder, GETTEXT_PACKAGE);
 
+	/* GtkSourceCompletionContainer is a private type. */
+	gtk_builder_expose_object (builder, "completion_container", G_OBJECT (container));
+
 	gtk_builder_add_from_resource (builder,
 				       "/org/gnome/gtksourceview/ui/gtksourcecompletion.ui",
-				       NULL);
+				       &error);
+
+	if (error != NULL)
+	{
+		g_error ("Error while loading the completion UI: %s", error->message);
+	}
 
 	init_tree_view (completion, builder);
 	init_main_window (completion, builder);
 	init_info_window (completion);
 
 	g_object_unref (builder);
+	g_object_unref (container);
 }
 
 void
