@@ -48,20 +48,23 @@
  * </example>
  *
  * If the calltip is displayed on top of a certain widget, say a #GtkTextView,
- * you should hide it when the #GtkWidget::focus-out-event signal is emitted by
- * the #GtkTextView. You may also be interested by the
- * #GtkTextBuffer:cursor-position property (when its value is modified). If you
- * use the #GtkSourceCompletionInfo through the #GtkSourceCompletion machinery,
- * you don't need to worry about this.
+ * you should attach the calltip window to the #GtkTextView with
+ * gtk_window_set_attached_to().  By doing this, the calltip will be hidden when
+ * the #GtkWidget::focus-out-event signal is emitted by the #GtkTextView. You
+ * may also be interested by the #GtkTextBuffer:cursor-position property (when
+ * its value is modified). If you use the #GtkSourceCompletionInfo through the
+ * #GtkSourceCompletion machinery, you don't need to worry about this.
  */
 
 #include <gtksourceview/gtksourcecompletioninfo.h>
-#include "gtksourcecompletionutils.h"
 #include "gtksourceview-i18n.h"
 
 struct _GtkSourceCompletionInfoPrivate
 {
 	guint idle_resize;
+
+	GtkWidget *attached_to;
+	gulong focus_out_event_handler;
 };
 
 /* Signals */
@@ -184,11 +187,55 @@ gtk_source_completion_info_get_preferred_height (GtkWidget *widget,
 
 /* Init, dispose, finalize, ... */
 
+static gboolean
+focus_out_event_cb (GtkSourceCompletionInfo *info)
+{
+	gtk_widget_hide (GTK_WIDGET (info));
+	return FALSE;
+}
+
+static void
+update_attached_to (GtkSourceCompletionInfo *info)
+{
+	GtkWidget *attached_to = gtk_window_get_attached_to (GTK_WINDOW (info));
+
+	if (info->priv->focus_out_event_handler != 0 &&
+	    info->priv->attached_to != NULL)
+	{
+		g_signal_handler_disconnect (info->priv->attached_to,
+					     info->priv->focus_out_event_handler);
+
+		info->priv->focus_out_event_handler = 0;
+	}
+
+	info->priv->attached_to = attached_to;
+
+	if (attached_to == NULL)
+	{
+		return;
+	}
+
+	g_object_add_weak_pointer (G_OBJECT (attached_to),
+				   (gpointer) &info->priv->attached_to);
+
+	info->priv->focus_out_event_handler = g_signal_connect_swapped (attached_to,
+									"focus-out-event",
+									G_CALLBACK (focus_out_event_cb),
+									info);
+}
+
 static void
 gtk_source_completion_info_init (GtkSourceCompletionInfo *info)
 {
 	info->priv = GTK_SOURCE_COMPLETION_INFO_GET_PRIVATE (info);
 
+	g_signal_connect (info,
+			  "notify::attached-to",
+			  G_CALLBACK (update_attached_to),
+			  NULL);
+
+	update_attached_to (info);
+
 	/* Tooltip style */
 	gtk_window_set_title (GTK_WINDOW (info), _("Completion Info"));
 	gtk_widget_set_name (GTK_WIDGET (info), "gtk-tooltip");
@@ -209,6 +256,13 @@ gtk_source_completion_info_finalize (GObject *object)
 		g_source_remove (info->priv->idle_resize);
 	}
 
+	if (info->priv->focus_out_event_handler != 0 &&
+	    info->priv->attached_to != NULL)
+	{
+		g_signal_handler_disconnect (info->priv->attached_to,
+					     info->priv->focus_out_event_handler);
+	}
+
 	G_OBJECT_CLASS (gtk_source_completion_info_parent_class)->finalize (object);
 }
 
@@ -260,11 +314,13 @@ gtk_source_completion_info_class_init (GtkSourceCompletionInfoClass *klass)
 	 * This signal is emitted before any "show" management. You can connect
 	 * to this signal if you want to change some properties or position
 	 * before the real "show".
+	 *
+	 * Deprecated: 3.10: This signal should not be used.
 	 */
 	signals[BEFORE_SHOW] =
 		g_signal_new ("before-show",
 		              G_TYPE_FROM_CLASS (klass),
-		              G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
+		              G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION | G_SIGNAL_DEPRECATED,
 		              0,
 		              NULL,
 		              NULL,
@@ -275,6 +331,176 @@ gtk_source_completion_info_class_init (GtkSourceCompletionInfoClass *klass)
 	g_type_class_add_private (object_class, sizeof (GtkSourceCompletionInfoPrivate));
 }
 
+/* Move to iter */
+
+static void
+get_iter_pos (GtkTextView *text_view,
+              GtkTextIter *iter,
+              gint        *x,
+              gint        *y,
+              gint        *height)
+{
+	GdkWindow *win;
+	GdkRectangle location;
+	gint win_x;
+	gint win_y;
+	gint xx;
+	gint yy;
+
+	gtk_text_view_get_iter_location (text_view, iter, &location);
+
+	gtk_text_view_buffer_to_window_coords (text_view,
+					       GTK_TEXT_WINDOW_WIDGET,
+					       location.x,
+					       location.y,
+					       &win_x,
+					       &win_y);
+
+	win = gtk_text_view_get_window (text_view, GTK_TEXT_WINDOW_WIDGET);
+	gdk_window_get_origin (win, &xx, &yy);
+
+	*x = win_x + xx;
+	*y = win_y + yy + location.height;
+	*height = location.height;
+}
+
+static void
+compensate_for_gravity (GtkSourceCompletionInfo *window,
+                        gint                    *x,
+                        gint                    *y,
+                        gint                     w,
+                        gint                     h)
+{
+	GdkGravity gravity = gtk_window_get_gravity (GTK_WINDOW (window));
+
+	/* Horizontal */
+	switch (gravity)
+	{
+		case GDK_GRAVITY_NORTH:
+		case GDK_GRAVITY_SOUTH:
+		case GDK_GRAVITY_CENTER:
+			*x = w / 2;
+			break;
+		case GDK_GRAVITY_NORTH_EAST:
+		case GDK_GRAVITY_SOUTH_EAST:
+		case GDK_GRAVITY_EAST:
+			*x = w;
+			break;
+		default:
+			*x = 0;
+			break;
+	}
+
+	/* Vertical */
+	switch (gravity)
+	{
+		case GDK_GRAVITY_WEST:
+		case GDK_GRAVITY_CENTER:
+		case GDK_GRAVITY_EAST:
+			*y = w / 2;
+			break;
+		case GDK_GRAVITY_SOUTH_EAST:
+		case GDK_GRAVITY_SOUTH:
+		case GDK_GRAVITY_SOUTH_WEST:
+			*y = w;
+			break;
+		default:
+			*y = 0;
+			break;
+	}
+}
+
+static void
+move_overlap (gint     *y,
+              gint      h,
+              gint      oy,
+              gint      cy,
+              gint      line_height,
+              gboolean  move_up)
+{
+	/* Test if there is overlap */
+	if (*y - cy < oy && *y - cy + h > oy - line_height)
+	{
+		if (move_up)
+		{
+			*y = oy - line_height - h + cy;
+		}
+		else
+		{
+			*y = oy + cy;
+		}
+	}
+}
+
+static void
+move_to_iter (GtkSourceCompletionInfo *window,
+	      GtkTextView             *view,
+	      GtkTextIter             *iter)
+{
+	GdkScreen *screen;
+	gint x, y;
+	gint w, h;
+	gint sw, sh;
+	gint cx, cy;
+	gint oy;
+	gint height;
+	gboolean overlapup;
+
+	screen = gtk_window_get_screen (GTK_WINDOW (window));
+
+	sw = gdk_screen_get_width (screen);
+	sh = gdk_screen_get_height (screen);
+
+	get_iter_pos (view, iter, &x, &y, &height);
+	gtk_window_get_size (GTK_WINDOW (window), &w, &h);
+
+	oy = y;
+	compensate_for_gravity (window, &cx, &cy, w, h);
+
+	/* Push window inside screen */
+	if (x - cx + w > sw)
+	{
+		x = (sw - w) + cx;
+	}
+	else if (x - cx < 0)
+	{
+		x = cx;
+	}
+
+	if (y - cy + h > sh)
+	{
+		y = (sh - h) + cy;
+		overlapup = TRUE;
+	}
+	else if (y - cy < 0)
+	{
+		y = cy;
+		overlapup = FALSE;
+	}
+	else
+	{
+		overlapup = TRUE;
+	}
+
+	/* Make sure that text is still readable */
+	move_overlap (&y, h, oy, cy, height, overlapup);
+
+	gtk_window_move (GTK_WINDOW (window), x, y);
+}
+
+static void
+move_to_cursor (GtkSourceCompletionInfo *window,
+		GtkTextView             *view)
+{
+	GtkTextBuffer *buffer;
+	GtkTextIter insert;
+
+	buffer = gtk_text_view_get_buffer (view);
+	gtk_text_buffer_get_iter_at_mark (buffer, &insert, gtk_text_buffer_get_insert (buffer));
+
+	move_to_iter (window, view, &insert);
+}
+
 /* Public functions */
 
 /**
@@ -307,18 +533,15 @@ gtk_source_completion_info_move_to_iter (GtkSourceCompletionInfo *info,
                                          GtkTextIter             *iter)
 {
 	g_return_if_fail (GTK_SOURCE_IS_COMPLETION_INFO (info));
-	g_return_if_fail (GTK_SOURCE_IS_VIEW (view));
+	g_return_if_fail (GTK_IS_TEXT_VIEW (view));
 
 	if (iter == NULL)
 	{
-		gtk_source_completion_utils_move_to_cursor (GTK_WINDOW (info),
-							    GTK_SOURCE_VIEW (view));
+		move_to_cursor (info, view);
 	}
 	else
 	{
-		gtk_source_completion_utils_move_to_iter (GTK_WINDOW (info),
-							  GTK_SOURCE_VIEW (view),
-							  iter);
+		move_to_iter (info, view, iter);
 	}
 }
 
